"""Psyrun specific exceptions and warnings."""

import warnings


class TaskWarning(UserWarning):
    """General warning related to task processing.

    *TaskWarnings* will always be shown by default.
    """
    def __init__(self, task_name, *args, **kwargs):
        super(TaskWarning, self).__init__(*args, **kwargs)
        self.task_name = task_name


class JobsRunningWarning(TaskWarning):
    """Warning issued when jobs for a task are still running."""
    def __init__(self, task_name):
        super(JobsRunningWarning, self).__init__(
            task_name,
            "Task '{}' has unfinished jobs queued. Not starting new jobs "
            "until these are finished or have been killed.".format(task_name))


class TaskWorkdirDirtyWarning(TaskWarning):
    """Warning issued when the workdir is dirty and would be overwritten."""
    def __init__(self, task_name):
        # TODO explain how to solve this
        super(TaskWorkdirDirtyWarning, self).__init__(
            task_name,
            "Work directory of task '{}' is dirty.".format(task_name))


class IneffectiveExcludeWarning(UserWarning):
    """Warning issued when a key in *exclude_from_result* was not found in the
    result.
    """
    def __init__(self, key):
        super(IneffectiveExcludeWarning, self).__init__(
            "The key '{}' to remove from the result was not found in the "
            "result dictionary.".format(key))
        self.key = key


warnings.simplefilter('always', category=TaskWarning)



import os
import os.path

import pytest

from psyrun.store import DefaultStore
from psyrun.backend.distribute import DistributeBackend, Splitter, Worker
from psyrun.jobs import Fullname, Submit, Uptodate
from psyrun.pspace import Param
from psyrun.tasks import TaskDef
from psyrun.utils.testing import MockScheduler, taskenv


def setup_fn(proc_id):
    assert proc_id == 0
    return {'p': 2}

def square(a, p):
    return {'x': a ** p}


@pytest.mark.parametrize(
    'pspace_size,max_splits,min_items,n_splits', [
        (7, 4, 4, 2), (8, 4, 4, 2), (9, 4, 4, 3),
        (15, 4, 4, 4), (16, 4, 4, 4), (17, 4, 4, 4),
        (15, 2, 4, 2), (16, 4, 16, 1)
    ])
class TestSplitter(object):
    def test_n_splits(
            self, tmpdir, pspace_size, max_splits, min_items, n_splits):
        splitter = Splitter(
            str(tmpdir), Param(x=range(pspace_size)), max_splits, min_items)
        assert splitter.n_splits == n_splits
        assert len(list(splitter.iter_in_out_files())) == n_splits

    def test_split_merge_roundtrip(
            self, tmpdir, pspace_size, max_splits, min_items, n_splits):
        splitter = Splitter(
            str(tmpdir), Param(x=range(pspace_size)), max_splits, min_items)
        splitter.split()

        for filename in os.listdir(splitter.indir):
            infile = os.path.join(splitter.indir, filename)
            outfile = os.path.join(splitter.outdir, filename)
            DefaultStore().save(outfile, DefaultStore().load(infile))

        result_file = os.path.join(str(tmpdir), 'result.npz')
        Splitter.merge(splitter.outdir, result_file)
        result = DefaultStore().load(result_file)
        assert sorted(result['x']) == sorted(range(pspace_size))


def test_worker(tmpdir):
    infile = str(tmpdir.join('in.npz'))
    outfile = str(tmpdir.join('out.npz'))
    DefaultStore().save(infile, Param(a=range(7)).build())
    worker = Worker(0, DefaultStore())
    worker.start(square, infile, outfile, setup_fn=setup_fn)
    result = DefaultStore().load(outfile)
    assert sorted(result['a']) == sorted(range(7))
    assert sorted(result['x']) == [i ** 2 for i in range(7)]


def test_get_queued(tmpdir, taskenv):
    task = TaskDef(os.path.join(taskenv.taskdir, 'task_square.py'))
    task.scheduler = MockScheduler(os.path.join(str(tmpdir), 'jobfile'))
    backend = DistributeBackend(task)
    job = backend.create_job()
    names = Fullname(job).names
    Submit(job, names, Uptodate(job, names, task))
    task.scheduler.consume_job(task.scheduler.joblist[0])
    task.scheduler.consume_job(task.scheduler.joblist[0])
    queued = backend.get_queued()
    assert sorted(queued.build()['x']) == [1, 2, 3]


def test_get_queued_without_split_completed(tmpdir, taskenv):
    task = TaskDef(os.path.join(taskenv.taskdir, 'task_square.py'))
    task.scheduler = MockScheduler(os.path.join(str(tmpdir), 'jobfile'))
    backend = DistributeBackend(task)
    job = backend.create_job()
    names = Fullname(job).names
    Submit(job, names, Uptodate(job, names, task))
    queued = backend.get_queued()
    assert sorted(queued.build()['x']) == [0, 1, 2, 3]


def test_get_failed_split(tmpdir, taskenv):
    task = TaskDef(os.path.join(taskenv.taskdir, 'task_square.py'))
    task.scheduler = MockScheduler(os.path.join(str(tmpdir), 'jobfile'))
    backend = DistributeBackend(task)
    job = backend.create_job()
    names = Fullname(job).names
    Submit(job, names, Uptodate(job, names, task))
    task.scheduler.kill(0)
    assert backend.get_failed() == ['square:split']


def test_get_failed_no_fails(tmpdir, taskenv):
    task = TaskDef(os.path.join(taskenv.taskdir, 'task_square.py'))
    task.scheduler = MockScheduler(os.path.join(str(tmpdir), 'jobfile'))
    backend = DistributeBackend(task)
    job = backend.create_job()
    names = Fullname(job).names
    Submit(job, names, Uptodate(job, names, task))
    assert backend.get_failed() == []
    task.scheduler.consume()
    assert backend.get_failed() == []


def test_get_failed_merge(tmpdir, taskenv):
    task = TaskDef(os.path.join(taskenv.taskdir, 'task_square.py'))
    task.scheduler = MockScheduler(os.path.join(str(tmpdir), 'jobfile'))
    backend = DistributeBackend(task)
    job = backend.create_job()
    names = Fullname(job).names
    Submit(job, names, Uptodate(job, names, task))
    while len(task.scheduler.joblist) > 1:
        task.scheduler.consume_job(task.scheduler.joblist[0])
    task.scheduler.kill(task.scheduler.joblist[0]['id'])
    assert backend.get_failed() == ['square:merge']


def test_get_failed_process(tmpdir, taskenv):
    task = TaskDef(os.path.join(taskenv.taskdir, 'task_square.py'))
    task.scheduler = MockScheduler(os.path.join(str(tmpdir), 'jobfile'))
    backend = DistributeBackend(task)
    job = backend.create_job()
    names = Fullname(job).names
    Submit(job, names, Uptodate(job, names, task))
    task.scheduler.consume_job(task.scheduler.joblist[0])
    task.scheduler.consume_job(task.scheduler.joblist[0])
    task.scheduler.kill(task.scheduler.joblist[0]['id'])
    assert backend.get_failed() == ['square:process:1']



import pytest

import numpy as np
from numpy.testing import assert_array_equal

from psyrun.store import AutodetectStore
from psyrun.store.h5 import H5Store
from psyrun.store.npz import NpzStore
from psyrun.store.pickle import PickleStore
from psyrun.pspace import Param


@pytest.mark.parametrize('store', [H5Store(), NpzStore(), PickleStore()])
class TestDictStores(object):
    def test_infile_with_none(self, store, tmpdir):
        filename = str(tmpdir.join('infile' + store.ext))
        pspace = (Param(a=[1.], b=[1.]) + Param(a=[2.], c=[2.])).build()
        store.save(filename, pspace)
        saved_pspace = store.load(filename)
        assert sorted(pspace.keys()) == sorted(saved_pspace.keys())
        for k in pspace:
            for a, b in zip(pspace[k], saved_pspace[k]):
                assert a == b or (np.isnan(a)and np.isnan(b))

    def test_load_specific_row(self, store, tmpdir):
        filename = str(tmpdir.join('infile' + store.ext))
        pspace = Param(a=range(5)).build()
        store.save(filename, pspace)
        saved_pspace = store.load(filename, row=2)
        assert sorted(pspace.keys()) == sorted(saved_pspace.keys())
        assert list(saved_pspace['a']) == [2]

    def test_merging_multidimensional_results(self, store, tmpdir):
        filename = str(tmpdir.join('r' + store.ext))

        data1 = {'a': np.zeros((2, 2, 2))}
        data2 = {'a': np.ones((3, 2, 2))}

        store.append(filename, data1)
        store.append(filename, data2)

        saved = store.load(filename)
        assert np.all(
            np.concatenate([data1['a'], data2['a']]) == np.asarray(saved['a']))

    def test_merging_results_with_varying_dimensionality(self, store, tmpdir):
        filename = str(tmpdir.join('r' + store.ext))

        data1 = {'a': np.zeros((1, 1, 2))}
        data2 = {'a': np.ones((2, 2, 1))}

        store.append(filename, data1)
        store.append(filename, data2)

        saved = store.load(filename)
        assert np.all(saved['a'][0][:1, :2] == data1['a'][0])
        assert np.all(saved['a'][1][:2, :1] == data2['a'][0])
        assert np.all(saved['a'][2][:2, :1] == data2['a'][1])


@pytest.mark.parametrize('store', [NpzStore(), PickleStore()])
class TestStringStoring(object):
    def test_saves_single_string_list(self, store, tmpdir):
        filename = str(tmpdir.join('r' + store.ext))

        data1 = {'a': ['A', 'B'], 'num': [1, 2]}
        data2 = {'a': ['AB'], 'num': [3]}

        store.append(filename, data1)
        store.append(filename, data2)

        saved = store.load(filename)
        assert np.all(np.concatenate([data1['a'], data2['a']]) == saved['a'])
        assert np.all(np.array([1, 2, 3]) == saved['num'])

    def test_handles_missing_values(self, store, tmpdir):
        filename = str(tmpdir.join('r' + store.ext))

        data1 = {'num': [1, 2], 'a': ['A', 'B']}
        data2 = {'num': [3], 'b': ['AB']}

        store.append(filename, data1)
        store.append(filename, data2)

        saved = store.load(filename)
        print(saved)
        assert np.all(saved['a'][:2] == data1['a'])
        assert np.all(saved['b'][2:] == data2['b'])


@pytest.mark.parametrize('store', [H5Store()])
def test_non_string_supporting(tmpdir, store):
    filename = str(tmpdir.join('r' + store.ext))

    with pytest.raises(NotImplementedError):
        store.save(filename, {'a': ['some string']})

    with pytest.raises(NotImplementedError):
        store.append(filename, {'a': ['some string']})


@pytest.mark.parametrize('store', [NpzStore(), PickleStore()])
def test_dict_and_lists(store, tmpdir):
    filename = str(tmpdir.join('r' + store.ext))

    obj1 = {'x': [1, 2]}
    obj2 = {'y': [3, 4]}

    store.append(filename, {'a': [obj1]})
    store.append(filename, {'a': [obj2]})

    saved = store.load(filename)
    print(store, saved)
    assert np.all(saved['a'] == [obj1, obj2])


class ObjectMock(object):
    def __init__(self, value):
        self.value = value


@pytest.mark.parametrize('store', [NpzStore(), PickleStore()])
def test_object(store, tmpdir):
    filename = str(tmpdir.join('r' + store.ext))

    obj1 = ObjectMock(1)
    obj2 = ObjectMock(2)

    store.append(filename, {'a': [obj1]})
    store.append(filename, {'a': [obj2]})

    saved = store.load(filename)
    print(store, saved)
    assert saved['a'][0].value == 1
    assert saved['a'][1].value == 2


@pytest.mark.parametrize('store', [NpzStore(), PickleStore(), H5Store()])
def test_autodetect_store_load(store, tmpdir):
    filename = str(tmpdir.join('r' + store.ext))
    store.save(filename, {'a': [1., 2., 3.]})

    data = AutodetectStore().load(filename)
    assert list(data.keys()) == ['a']
    assert_array_equal(data['a'], [1., 2., 3.])


@pytest.mark.parametrize('store', [NpzStore(), PickleStore(), H5Store()])
def test_autodetect_store_save(store, tmpdir):
    filename = str(tmpdir.join('r' + store.ext))
    AutodetectStore().save(filename, {'a': [1., 2., 3.]})

    data = store.load(filename)
    assert list(data.keys()) == ['a']
    assert_array_equal(data['a'], [1., 2., 3.])


@pytest.mark.parametrize('store', [NpzStore(), PickleStore(), H5Store()])
def test_autodetect_store_append(store, tmpdir):
    filename = str(tmpdir.join('r' + store.ext))
    store.append(filename, {'a': [1.]})
    AutodetectStore().append(filename, {'a': [2., 3.]})

    data = store.load(filename)
    assert list(data.keys()) == ['a']
    assert_array_equal(data['a'], [1., 2., 3.])



"""Job scheduler."""

from collections import namedtuple
import getpass
import os
import os.path
import re
import subprocess
import sys


JobStatus = namedtuple('JobStatus', ['id', 'status', 'name'])


class Scheduler(object):
    """Scheduler interface."""

    USER_DEFAULT_ARGS = {}

    def submit(
            self, args, output_filename, name=None, depends_on=None,
            scheduler_args=None):
        """Submit a job.

        Parameters
        ----------
        args : sequence
            The command and arguments to execute.
        output_filename : str
            File to write process output to.
        name : str, optional
            Name of job.
        depends_on : sequence of int, optional
            IDs of jobs that need to finish first before the submitted job can
            be started.
        scheduler_args : dict, optional
            Additional arguments for the scheduler.

        Returns
        -------
        Job ID
        """
        raise NotImplementedError()

    def submit_array(
            self, n, args, output_filename, name=None, depends_on=None,
            scheduler_args=None):
        """Submit a job array.

        If the scheduler does not support job arrays, this method should raise
        *NotImplementedError*.

        Parameters
        ----------
        n : int
            Number of tasks to submit.
        args : sequence
            The command and arguments to execute. The string ``'%a'`` will be
            replaced with the task number in each argument.
        output_filename : str
            File to write process output to. The string ``'%a'`` will be
            replaced by the task number.
        name : str, optional
            Name of job.
        depends_on : sequence of int, optional
            IDs of jobs that need to finish first before the submitted job can
            be started.
        scheduler_args : dict, optional
            Additional arguments for the scheduler.

        Returns
        -------
        Job ID
        """
        raise NotImplementedError()

    def kill(self, jobid):
        """Kill a job.

        Parameters
        ----------
        jobid
            Job to kill.
        """
        raise NotImplementedError()

    def get_status(self, jobid):
        """Get the status of a job.

        Parameters
        ----------
        jobid
            Job to request status of.

        Returns
        -------
        `JobStatus`
            Returns a tuple with *(id, status, name)* wherein status can be

            * ``'Q'`` for a queued job
            * ``'*Q'`` for a queued job waiting on another job to finish
            * ``'Z'`` for a sleeping job
            * ``'D'`` for a completed job

            If no status data is available for the job ID, None will be
            returned.
        """
        raise NotImplementedError()

    def get_jobs(self):
        """Get all queued, running, and recently finished jobs.

        Returns
        -------
        sequence
            Job IDs
        """
        raise NotImplementedError()


class ImmediateRun(Scheduler):
    """Runs jobs immediately on the local machine."""

    _cur_id = 0
    _failed_jobs = []

    def submit(
            self, args, output_filename, name=None, depends_on=None,
            scheduler_args=None):
        """Submit a job.

        Parameters
        ----------
        args : sequence
            The command and arguments to execute.
        output_filename : str
            File to write process output to.
        name
            Unused.
        depends_on
            Unused
        scheduler_args
            Unused.

        Returns
        -------
        int
            Job ID
        """
        if depends_on is None:
            depends_on = []

        self._cur_id += 1

        if any(x in self._failed_jobs for x in depends_on):
            self._failed_jobs.append(self._cur_id)
            return self._cur_id

        with open(output_filename, 'a') as f:
            if subprocess.call(args, stdout=f, stderr=subprocess.STDOUT) != 0:
                self._failed_jobs.append(self._cur_id)
        return self._cur_id

    def submit_array(
            self, n, args, output_filename, name=None, depends_on=None,
            scheduler_args=None):
        raise NotImplementedError("Job arrays not supported.")

    def kill(self, jobid):
        """Has no effect."""
        pass

    def get_status(self, jobid):
        """Has no effect."""
        pass

    def get_jobs(self):
        """Returns an empty list."""
        return []


class ExternalScheduler(Scheduler):
    class Option(object):
        def __init__(self, name, conversion=str):
            self.name = name
            self.conversion = conversion

        def build(self, value):
            raise NotImplementedError()

        def __repr__(self):
            return "<{clsname} '{name}'>".format(
                clsname=self.__class__.__name__, name=self.name)

    class ShortOption(Option):
        def build(self, value):
            if value is None:
                return []
            return [self.name, self.conversion(value)]

    class LongOption(Option):
        def build(self, value):
            if value is None:
                return []
            return [self.name + '=' + self.conversion(value)]

    KNOWN_ARGS = {}

    def build_args(self, **kwargs):
        args = []
        for k, v in kwargs.items():
            args.extend(self.KNOWN_ARGS[k].build(v))
        return args


class Sqsub(ExternalScheduler):
    """sqsub (sharcnet) scheduler."""

    USER_DEFAULT_ARGS = {
        'timelimit': '1h',
        'n_cpus': 1,
        'n_nodes': 1,
        'memory': '1G',
    }

    KNOWN_ARGS = {
        'timelimit': ExternalScheduler.ShortOption('-r'),
        'output_file': ExternalScheduler.ShortOption('-o'),
        'n_cpus': ExternalScheduler.ShortOption('-n'),
        'n_nodes': ExternalScheduler.ShortOption('-N'),
        'memory': ExternalScheduler.LongOption('--mpp'),
        'depends_on': ExternalScheduler.ShortOption(
            '-w', lambda jobids: ','.join(str(x) for x in jobids)),
        'idfile': ExternalScheduler.LongOption('--idfile'),
        'name': ExternalScheduler.ShortOption('-j')
    }

    def __init__(self, workdir=None):
        super(Sqsub, self).__init__()

        if workdir is None:
            workdir = '/work/{user}/psyrun'.format(user=os.environ['USER'])
        if not os.path.exists(workdir):
            os.makedirs(workdir)
        self.workdir = workdir
        self.idfile = os.path.join(workdir, 'idfile')
        self._jobs = None

    def submit(
            self, args, output_filename, name=None, depends_on=None,
            scheduler_args=None):
        """Submit a job.

        Parameters
        ----------
        args : list
            The command and arguments to execute.
        output_filename : str
            File to write process output to.
        name : str, optional
            Name of job.
        depends_on : list of int, optional
            IDs of jobs that need to finish first before the submitted job can
            be started.
        scheduler_args : dict, optional
            Additional arguments for the scheduler.

        Returns
        -------
        int
            Job ID
        """
        if scheduler_args is None:
            scheduler_args = dict()
        else:
            scheduler_args = {k: v(name) if callable(v) else v
                              for k, v in scheduler_args.items()}

        # Checking whether jobs depending on are completed before submitting
        # the new job is a race condition, but I see no possibility to avoid
        # it.
        if depends_on is not None:
            statii = [self.get_status(x) for x in depends_on]
            depends_on = [x for x, s in zip(depends_on, statii)
                          if s is not None and s.status != 'D']

        scheduler_args.update({
            'idfile': self.idfile,
            'output_file': output_filename,
            'depends_on': depends_on,
            'name': name,
        })
        subprocess.check_call(
            ['sqsub'] + self.build_args(**scheduler_args) + args)
        with open(self.idfile, 'r') as f:
            return int(f.read())

    def submit_array(
            self, n, args, output_filename, name=None, depends_on=None,
            scheduler_args=None):
        raise NotImplementedError("Job arrays not supported.")

    def kill(self, jobid):
        """Kill a job.

        Parameters
        ----------
        jobid : int
            Job to kill.
        """
        subprocess.check_call(['sqkill', str(jobid)])

    def get_status(self, jobid=None):
        """Get the status of a job.

        Parameters
        ----------
        jobid : int
            Job to request status of.

        Returns
        -------
        namedtuple or None
            Returns a tuple with *(id, status, name)* wherein status can be

            - ``'R'`` for a running job
            - ``'Q'`` for a queued job
            - ``'*Q'`` for a queued job waiting on another job to finish
            - ``'Z'`` for a sleeping job
            - ``'D'`` for a completed job

            If no status data is available for the job ID, None will be
            returned.
        """
        if self._jobs is None:
            self.refresh_job_info()
        elif jobid not in self._jobs:
            try:
                stdout = subprocess.check_output(
                    ['sqjobs', str(jobid)], stderr=subprocess.STDOUT,
                    universal_newlines=True)
                for line in stdout.split('\n')[2:]:
                    cols = line.split(None, 5)
                    if len(cols) > 3 and int(cols[0]) == jobid:
                        if cols[2] == 'C':
                            cols[2] = 'D'
                        self._jobs[jobid] = JobStatus(jobid, cols[2], cols[-1])
            except subprocess.CalledProcessError as err:
                # 1 if none pending, running, or suspended
                if err.returncode != 1:
                    raise
        return self._jobs.get(jobid, None)

    def get_jobs(self):
        """Get all queued and running jobs.

        Returns
        -------
        list of int
            Job IDs
        """
        if self._jobs is None:
            self.refresh_job_info()
        return self._jobs.keys()

    def refresh_job_info(self):
        self._jobs = {}
        stdout = subprocess.check_output(['sqjobs'], universal_newlines=True)
        for line in stdout.split('\n')[2:]:
            cols = line.split(None, 5)
            if len(cols) > 2 and cols[2] in ['R', 'Q', '*Q', 'Z']:
                jobid = int(cols[0])
                self._jobs[jobid] = JobStatus(jobid, cols[2], cols[-1])


class Slurm(ExternalScheduler):
    """Slurm (sbatch) scheduler."""

    USER_DEFAULT_ARGS = {
        'timelimit': '1h',
        'memory': '1G',
    }

    KNOWN_ARGS = {
        'timelimit': ExternalScheduler.ShortOption('-t'),
        'output_file': ExternalScheduler.ShortOption('-o'),
        'cores-per-socket': ExternalScheduler.LongOption('--cores-per-socket'),
        'sockets-per-node': ExternalScheduler.LongOption('--sockets-per-node'),
        'n_cpus': ExternalScheduler.ShortOption('-c'),
        'n_nodes': ExternalScheduler.ShortOption('-N'),
        'memory': ExternalScheduler.LongOption('--mem'),
        'memory_per_cpu': ExternalScheduler.LongOption('--mem-per-cpu'),
        'depends_on': ExternalScheduler.ShortOption(
            '-d', lambda jobids: 'afterok:' + ':'.join(
                str(x) for x in jobids)),
        'name': ExternalScheduler.ShortOption('-J'),
        'array': ExternalScheduler.LongOption('--array'),
    }

    STATUS_MAP = {
        'PD': 'Q',
        'R': 'R',
        'CA': 'D',
        'CF': 'R',
        'CG': 'R',
        'CD': 'D',
        'F': 'D',
        'TO': 'D',
        'NF': 'D',
        'RV': 'D',
        'SE': 'D',
    }

    def __init__(self, workdir=None):
        super(Slurm, self).__init__()
        if workdir is None:
            workdir = '/project/{user}/psyrun'.format(user=os.environ['USER'])
        if not os.path.exists(workdir):
            os.makedirs(workdir)
        self.workdir = workdir
        self._jobs = None

    def submit(
            self, args, output_filename, name=None, depends_on=None,
            scheduler_args=None):
        """Submit a job.

        Parameters
        ----------
        args : list
            The command and arguments to execute.
        output_filename : str
            File to write process output to.
        name : str, optional
            Name of job.
        depends_on : list of int, optional
            IDs of jobs that need to finish first before the submitted job can
            be started.
        scheduler_args : dict, optional
            Additional arguments for the scheduler.

        Returns
        -------
        str
            Job ID
        """
        out = subprocess.check_output(self._prepare_submisson_cmd(
            args, output_filename, name, scheduler_args, depends_on))
        out = out.decode('utf-8')
        return out.split(None)[-1]

    def submit_array(
            self, n, args, output_filename, name=None, depends_on=None,
            scheduler_args=None):
        if scheduler_args is None:
            scheduler_args = dict()
        else:
            scheduler_args = dict(scheduler_args)

        scheduler_args['array'] = '0-' + str(n - 1)

        code = '''#!{python}
import os
import subprocess
import sys

task_id = os.environ['SLURM_ARRAY_TASK_ID']
sys.exit(subprocess.call([a.replace('%a', str(task_id)) for a in {args!r}]))
'''.format(python=sys.executable, args=args)

        cmd = self._prepare_submisson_cmd(
            [], output_filename, name, scheduler_args, depends_on)
        p = subprocess.Popen(
            cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        out, _ = p.communicate(code.encode('utf-8'))
        out = out.decode('utf-8')
        if p.returncode != 0:
            raise subprocess.CalledProcessError(p.returncode, cmd, out)
        return out.split(None)[-1]

    def _prepare_submisson_cmd(
            self, args, output_filename, name=None, scheduler_args=None,
            depends_on=None):
        if scheduler_args is None:
            scheduler_args = dict()
        else:
            scheduler_args = {k: v(name) if callable(v) else v
                              for k, v in scheduler_args.items()}

        # Checking whether jobs depending on are completed before submitting
        # the new job is a race condition, but I see no possibility to avoid
        # it.
        if depends_on is not None:
            statii = [self.get_status(x) for x in depends_on]
            depends_on = [x for x, s in zip(depends_on, statii)
                          if s is not None and s.status != 'D']

        scheduler_args.update({
            'output_file': output_filename,
            'name': name,
        })
        if depends_on is not None and len(depends_on) > 0:
            scheduler_args['depends_on'] = depends_on
        return ['sbatch'] + self.build_args(**scheduler_args) + args

    def kill(self, jobid):
        """Kill a job.

        Parameters
        ----------
        jobid : str
            Job to kill.
        """
        subprocess.check_call(['scancel', str(jobid)])

    def get_status(self, jobid=None):
        """Get the status of a job.

        Parameters
        ----------
        jobid : str
            Job to request status of.

        Returns
        -------
        namedtuple or None
            Returns a tuple with *(id, status, name)* wherein status can be

            - ``'R'`` for a running job
            - ``'Q'`` for a queued job
            - ``'*Q'`` for a queued job waiting on another job to finish
            - ``'Z'`` for a sleeping job
            - ``'D'`` for a completed job

            If no status data is available for the job ID, None will be
            returned.
        """
        if self._jobs is None:
            self.refresh_job_info()
        elif jobid not in self._jobs:
            stdout = subprocess.check_output(
                ['squeue', '-u', getpass.getuser(), '-h', '-o', '%A %t %j %r',
                 '-j', str(jobid)],
                stderr=subprocess.STDOUT, universal_newlines=True)
            self._update_jobs(stdout)
        return self._jobs.get(jobid, None)

    def get_jobs(self):
        """Get all queued and running jobs.

        Returns
        -------
        list of str
            Job IDs
        """
        if self._jobs is None:
            self.refresh_job_info()
        return self._jobs.keys()

    def refresh_job_info(self):
        self._jobs = {}
        stdout = subprocess.check_output(
            ['squeue', '-u', getpass.getuser(), '-h', '-o', '%i %t %j %r'],
            universal_newlines=True)
        self._update_jobs(stdout)

    def _update_jobs(self, squeue_out):
        for line in squeue_out.split('\n'):
            cols = line.split(None, 4)
            if len(cols) > 3 and cols[1] in ['PD', 'R', 'CF', 'CG']:
                jobid = cols[0]
                status = self.STATUS_MAP[cols[1]]
                if status == 'Q' and cols[3] == 'Dependency':
                    status = '*Q'
                m = re.match(r'^(\d+)_\[(\d+)-(\d+)\]$', jobid)
                if m:
                    for i in range(int(m.group(2)), int(m.group(3)) + 1):
                        sub_id = m.group(1) + '_' + str(i)
                        self._jobs[sub_id] = JobStatus(
                            sub_id, status, cols[2] + ':' + str(i))
                else:
                    self._jobs[jobid] = JobStatus(jobid, status, cols[2])



import numpy as np
from numpy.testing import assert_equal
import pytest

from psyrun.store.np_shape import min_shape, match_shape


class Obj(object):
    def __init__(self, value):
        self.value = value

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return isinstance(other, Obj) and self.value == other.value


def test_min_shape():
    assert min_shape(((1, 1, 1), (2, 3, 4))) == (2, 3, 4)
    assert min_shape(((3, 2, 1), (1, 2, 3))) == (3, 2, 3)
    assert min_shape(((1, 1, 2), (1, 2, 1), (2, 1, 1))) == (2, 2, 2)
    assert min_shape(((1,), (2,))) == (2,)
    assert min_shape((tuple(), tuple())) == tuple()
    with pytest.raises(ValueError):
        assert min_shape(((1, 3), (2, 1, 3)))


@pytest.mark.parametrize('a,shape,expected', [
    ([[1.], [2.]], (2,), [[1., np.nan], [2., np.nan]]),
    ([[1j], [2j]], (2,), [[1j, np.nan], [2j, np.nan]]),
    ([[1], [2]], (2,), [[1., np.nan], [2., np.nan]]),
    ([['a'], ['b']], (2,), [['a', ''], ['b', '']]),
    ([[Obj(1)], [Obj(2)]], (2,), [[Obj(1), None], [Obj(2), None]]),
    (1., (1,), [[1.]]),
])
def test_match_shape(a, shape, expected):
    assert_equal(match_shape(a, shape), np.asarray(expected))


def test_match_shape_with_unsupported_missing():
    with pytest.raises(ValueError):
        match_shape(np.ones((2, 1), dtype='M'), (2,))



import subprocess
import sys

from psyrun.scheduler import ImmediateRun, JobStatus, Slurm


def test_immediate_run_submit(tmpdir):
    outfile = str(tmpdir) + '/out'
    sched = ImmediateRun()
    sched.submit(['echo', 'success'], outfile)
    with open(outfile, 'r') as f:
        assert f.read().strip() == 'success'


def test_immediate_run_failing_job(tmpdir):
    outfile = str(tmpdir) + '/out'
    sched = ImmediateRun()
    jobid = sched.submit(['false'], outfile)
    jobid = sched.submit(['echo', 'not echoed'], outfile, depends_on=[jobid])

    with open(outfile, 'r') as f:
        assert f.read().strip() == ''


def test_slurm_submit(monkeypatch, tmpdir):
    def check_output(cmd, universal_newlines=None):
        assert sorted(cmd) == sorted(
            ['sbatch', '-o', 'outfile', '-J', 'name', '0', '1'])
        return b'name 1'

    monkeypatch.setattr(subprocess, 'check_output', check_output)

    sched = Slurm(str(tmpdir))
    assert sched.submit(['0', '1'], 'outfile', name='name') == '1'


def test_slurm_submit_with_additional_args(monkeypatch, tmpdir):
    def check_output(cmd, universal_newlines=None):
        if cmd[0] == 'squeue':
            return '\n0 R name None\n'
        assert sorted(cmd) == sorted([
            'sbatch', '-t', '1h', '-o', 'outfile', '-J', 'name', '-d',
            'afterok:0', '0', '1'])
        return b'name 1'

    monkeypatch.setattr(subprocess, 'check_output', check_output)

    sched = Slurm(str(tmpdir))
    assert sched.submit(
        ['0', '1'], 'outfile', name='name', depends_on=['0'],
        scheduler_args={'timelimit': '1h'}) == '1'


def test_slurm_kill(monkeypatch, tmpdir):
    def check_call(cmd):
        assert cmd == ['scancel', '1']

    monkeypatch.setattr(subprocess, 'check_call', check_call)

    sched = Slurm(str(tmpdir))
    sched.kill('1')


def test_slurm_status(monkeypatch, tmpdir):
    def check_output(cmd, universal_newlines=None, stderr=None):
        assert cmd[:2] == ['squeue', '-u']
        assert cmd[3:5] == ['-h', '-o']
        assert cmd[5] == '%i %t %j %r' or cmd[5] == '%A %t %j %r'
        return '''id state name reason
1 R running None
2 PD pending Dependency
3 PD priority Priority
4_[0-2] PD array Priority
'''

    monkeypatch.setattr(subprocess, 'check_output', check_output)
    sched = Slurm(str(tmpdir))
    assert sorted(sched.get_jobs()) == ['1', '2', '3', '4_0', '4_1', '4_2']
    assert sched.get_status('1') == JobStatus('1', 'R', 'running')
    assert sched.get_status('2') == JobStatus('2', '*Q', 'pending')
    assert sched.get_status('3') == JobStatus('3', 'Q', 'priority')
    for i in range(3):
        job_id = '4_' + str(i)
        assert sched.get_status(job_id) == JobStatus(
            job_id, 'Q', 'array:' + str(i))


def test_slurm_submit_array(monkeypatch, tmpdir):
    class Popen(object):
        def __init__(self, cmd, stdin=None, stdout=None):
            assert sorted(cmd) == sorted(
                ['sbatch', '--array=0-2', '-o', 'outfile'])
            self.returncode = 0

        def communicate(self, input):
            assert input.decode('utf-8') == '#!' + sys.executable + '''
import os
import subprocess
import sys

task_id = os.environ['SLURM_ARRAY_TASK_ID']
sys.exit(subprocess.call([a.replace('%a', str(task_id)) for a in ['arg0', 'arg1']]))
'''
            return b'array 1', b''

    monkeypatch.setattr(subprocess, 'Popen', Popen)
    sched = Slurm(str(tmpdir))
    sched.submit_array(3, ['arg0', 'arg1'], 'outfile')



"""Base backend interface."""

import os
import stat
import sys


class JobSourceFile(object):
    """Describes a source code file for a job.

    Parameters
    ----------
    path : str
        Path to write the source file to.
    task : Task
        Task the source file corresponds to.
    job_code : str
        Job specific code to execute.

    Attributes
    ----------
    written : bool
        Whether the source file has been written.
    full_code : str
        The full job code including non job-specific parts that are used for
        every job.
    """

    def __init__(self, path, task, job_code):
        self.path = path
        self.task = task
        self.job_code = job_code
        self.written = False

    def write(self):
        """Write the job code to the file *self.path*."""
        with open(self.path, 'w') as f:
            f.write(self.full_code)
            fd = f.fileno()
            os.fchmod(fd, os.fstat(fd).st_mode | stat.S_IXUSR)
        self.written = True

    @property
    def full_code(self):
        return '''#!{python}
try:
    import faulthandler
    faulthandler.enable()
except:
    pass

import time
print("")
print("----------------------------------------------------------------------")
print("Job started ({{}})".format(time.strftime('%a, %d %b %Y %H:%M:%S, %Z')))
print("----------------------------------------------------------------------")

import os
os.chdir({taskdir!r})

from psyrun.tasks import TaskDef
task = TaskDef({taskpath!r})
{code}
        '''.format(
            python=self.task.python,
            path=sys.path,
            taskdir=os.path.abspath(os.path.dirname(self.task.path)),
            taskpath=os.path.abspath(self.task.path), code=self.job_code)


class Backend(object):
    """Abstract base class for processing backends.

    Processing backends determine how work is split across jobs.

    Deriving classes are supposed to implement `create_job` and `get_missing`.

    Parameters
    ----------
    task : `TaskDef`
        The task to create processing jobs for.

    Attributes
    ----------
    task : `TaskDef`
        The task to create processing jobs for.
    workdir : str
        Directory in which supporting files for processing the task are stored.
    """

    def __init__(self, task):
        super(Backend, self).__init__()
        self.task = task
        self.workdir = os.path.join(task.workdir, task.name)
        if not os.path.exists(self.workdir):
            os.makedirs(self.workdir)

    def submit_code(self, code, name, depends_on=None, args=None):
        """Submits some code to execute to the task scheduler.

        Parameters
        ----------
        code : str
            Code to execute in job.
        name : str
            Job name.
        depends_on : sequence
            Job IDs that have to finish before the submitted code can be
            executed.
        args : sequence
            Additional arguments to pass to the job.

        Returns
        -------
        dict
            Contains the id of the submitted job under the key ``'id'``.
        """
        codefile = os.path.join(self.workdir, name + '.py')
        return self.submit_file(JobSourceFile(codefile, self.task, code),
                                name, depends_on=depends_on, args=args)

    def submit_file(self, job_source_file, name, depends_on=None, args=None):
        """Submits a source file to execute to the task scheduler.

        Parameters
        ----------
        job_source_file : `JobSourceFile`
            Source file to execute in job.
        name: str
            Job name.
        depends_on: sequence
            Job IDs that have to finish before the submitted code can be
            executed.
        args : sequence
            Additional arguments to pass to the job.

        Returns
        -------
        dict
            Contains the id of the submitted job under the key ``'id'``.
        """
        if args is None:
            args = []

        self._prepare_job_submission(job_source_file, name)

        output_filename = os.path.join(self.workdir, name + '.log')
        return self.task.scheduler.submit(
            [job_source_file.path] + args, output_filename,
            name, depends_on, self.task.scheduler_args)

    def submit_array(
            self, n, job_source_file, name, depends_on=None, args=None):
        """Submits a source file to execute to the task scheduler.

        Parameters
        ----------
        job_source_file : `JobSourceFile`
            Source file to execute in job.
        name: str
            Job name.
        depends_on: sequence
            Job IDs that have to finish before the submitted code can be
            executed.
        args : sequence
            Additional arguments to pass to the job.

        Returns
        -------
        dict
            Contains the id of the submitted job under the key ``'id'``.
        """
        if args is None:
            args = []

        self._prepare_job_submission(job_source_file, name)

        output_filename = os.path.join(self.workdir, name + ':%a.log')
        return self.task.scheduler.submit_array(
            n, [job_source_file.path] + args, output_filename,
            name, depends_on, self.task.scheduler_args)

    def _prepare_job_submission(self, job_source_file, name):
        if not job_source_file.written:
            job_source_file.write()

        for job in self.task.scheduler.get_jobs():
            status = self.task.scheduler.get_status(job)
            if status is not None and name == status.name:
                self.task.scheduler.kill(job)

    def create_job(self, cont=False):
        """Create the job tree to process given task.

        Parameters
        ----------
        cont : bool, optional
            By default old results will be discarded, but when this option is
            set to True, old results will be kept and merged with the new
            results.
        """
        raise NotImplementedError()

    def get_missing(self):
        """Returns a `ParameterSpace` with missing parameter assignments.

        Missing paramaters assignments are parameter assignments requested by
        the task definition, but that have not been evaluated yet.
        """
        raise NotImplementedError

    def get_queued(self):
        """Returns parameter sets that are still in queue to be processed.

        May return ``None`` if this is not supported by the backend.
        """
        raise NotImplementedError()

    def get_failed(self):
        """Returns a list of failed jobs.

        May return ``None`` if this is not supported by the backend.
        """
        raise NotImplementedError()



from __future__ import unicode_literals

import os
import os.path
import re
import shutil
import time

import numpy as np
import pytest

from psyrun.main import psy_main
from psyrun.exceptions import JobsRunningWarning, TaskWorkdirDirtyWarning
from psyrun.store.h5 import H5Store
from psyrun.store.npz import NpzStore
from psyrun.store.pickle import PickleStore
from psyrun.tasks import TaskDef, Config
from psyrun.utils.testing import MockScheduler, TASKDIR, taskenv


@pytest.fixture
def scheduler(taskenv, request):
    jobfile = os.path.join(taskenv.rootdir, 'jobfile')
    mock = MockScheduler(jobfile)

    def fin():
        try:
            os.remove(jobfile)
        except OSError:
            pass

    request.addfinalizer(fin)
    return mock


def get_task_path(name):
    return os.path.join(TASKDIR, 'task_' + name + '.py')


def test_load_task_defaults():
    task = TaskDef(get_task_path('square'))
    assert task.path == get_task_path('square')
    assert task.name == 'square'
    assert hasattr(task, 'scheduler')
    assert hasattr(task, 'scheduler_args')
    assert hasattr(task, 'python')


def test_load_task_uses_config_as_default():
    conf = Config()
    conf.python = 'env python'
    task1 = TaskDef(get_task_path('square'), conf)
    assert task1.python == 'env python'
    task2 = TaskDef(get_task_path('noop'), conf)
    assert task2.python == 'true'


def test_load_config_from_file(tmpdir):
    conffile = os.path.join(str(tmpdir), 'conf.py')
    with open(conffile, 'w') as f:
        f.write('python = "env python"')
    conf = Config.load_from_file(conffile)
    assert conf.python == 'env python'


@pytest.mark.parametrize('task', ['square', 'square_load_balanced'])
def test_psyrun(taskenv, task):
    psy_main(['run', '--taskdir', taskenv.taskdir, task])
    result = PickleStore().load(
        os.path.join(taskenv.workdir, task, 'result.pkl'))
    assert sorted(result['y']) == [0, 1, 4, 9]


def test_psyrun_h5_backend(taskenv):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'square_h5'])
    result = H5Store().load(
        os.path.join(taskenv.workdir, 'square_h5', 'result.h5'))
    assert sorted(result['y']) == [0, 1, 4, 9]


def test_psyrun_npz_backend(taskenv):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'square_npz'])
    result = NpzStore().load(
        os.path.join(taskenv.workdir, 'square_npz', 'result.npz'))
    assert sorted(result['y']) == [0, 1, 4, 9]


@pytest.mark.parametrize('task', ['square', 'square_load_balanced'])
def test_exclude_from_result(taskenv, task):
    psy_main(['run', '--taskdir', taskenv.taskdir, task])
    result = PickleStore().load(
        os.path.join(taskenv.workdir, task, 'result.pkl'))
    assert 'z' not in result


def test_fails_for_existing_old_results_by_default(taskenv):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'square'])
    # Still up to date, not warning
    with pytest.warns(None) as record:
        psy_main(['run', '--taskdir', taskenv.taskdir, 'square'])
    for w in record:
        assert not issubclass(w.category, TaskWorkdirDirtyWarning)
    time.sleep(1)
    os.utime(os.path.join(taskenv.taskdir, 'task_square.py'), None)
    with pytest.warns(TaskWorkdirDirtyWarning):
        psy_main(['run', '--taskdir', taskenv.taskdir, 'square'])


@pytest.mark.parametrize('task', ['square', 'square_load_balanced'])
def test_psyrun_can_continue_interrupted_job(taskenv, task):
    psy_main(['run', '--taskdir', taskenv.taskdir, task])
    result = PickleStore().load(
        os.path.join(taskenv.workdir, task, 'result.pkl'))
    assert sorted(result['y']) == [0, 1, 4, 9]
    time.sleep(1)
    with open(os.path.join(
        taskenv.taskdir, 'task_{}.py'.format(task)), 'a') as f:
        f.write('\npspace += Param(x=[4, 5])\n')
    psy_main(['run', '--taskdir', taskenv.taskdir, '-c', task])
    result = PickleStore().load(
        os.path.join(taskenv.workdir, task, 'result.pkl'))
    assert sorted(result['y']) == [0, 1, 4, 9, 16, 25]


@pytest.mark.parametrize('task', ['square', 'square_load_balanced'])
def test_psyrun_can_continue_job_with_outdated_taskfile(taskenv, task):
    result_file = os.path.join(taskenv.workdir, task, 'result.pkl')
    psy_main(['run', '--taskdir', taskenv.taskdir, task])
    result = PickleStore().load(result_file)
    assert sorted(result['y']) == [0, 1, 4, 9]
    with open(os.path.join(
        taskenv.taskdir, 'task_{}.py'.format(task)), 'a') as f:
        f.write('\npspace += Param(x=[4, 5])\n')
    psy_main(['status', '--taskdir', taskenv.taskdir,  task])
    os.utime(result_file, None)
    psy_main(['run', '--taskdir', taskenv.taskdir, '-c', task])
    result = PickleStore().load(result_file)
    assert sorted(result['y']) == [0, 1, 4, 9, 16, 25]


def test_psyrun_can_continue_interrupted_job_no_result_file(taskenv):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'square'])
    result = PickleStore().load(
        os.path.join(taskenv.workdir, 'square', 'result.pkl'))
    assert sorted(result['y']) == [0, 1, 4, 9]
    with open(os.path.join(taskenv.taskdir, 'task_square.py'), 'a') as f:
        f.write('\npspace += Param(x=[4, 5])\n')
    os.remove(os.path.join(taskenv.workdir, 'square', 'result.pkl'))
    psy_main(['run', '--taskdir', taskenv.taskdir, '-c', 'square'])
    result = PickleStore().load(
        os.path.join(taskenv.workdir, 'square', 'result.pkl'))
    assert sorted(result['y']) == [0, 1, 4, 9, 16, 25]


def test_allows_to_clean_results(taskenv):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'square'])
    time.sleep(1)
    os.utime(os.path.join(taskenv.taskdir, 'task_square.py'), None)
    psy_main(['clean', '--taskdir', taskenv.taskdir, 'square'])
    with pytest.warns(None) as record:
        psy_main(['run', '--taskdir', taskenv.taskdir, 'square'])
    for w in record:
        assert not issubclass(w.category, TaskWorkdirDirtyWarning)


def test_psyrun_workdir_contents(taskenv):
    workdir = os.path.join('psy-work', 'square')
    os.remove(os.path.join(taskenv.taskdir, 'psy-conf.py'))
    psy_main(['run', '--taskdir', taskenv.taskdir, 'square'])
    assert os.path.exists(os.path.join(workdir, 'in', '0.pkl'))
    assert os.path.exists(os.path.join(workdir, 'out', '0.pkl'))
    assert os.path.exists(os.path.join(workdir, 'result.pkl'))
    assert os.path.exists(os.path.join(workdir, 'square:split.py'))
    assert os.path.exists(os.path.join(workdir, 'square:process.py'))
    assert os.path.exists(os.path.join(workdir, 'square:merge.py'))
    assert os.path.exists(os.path.join(workdir, 'square:split.log'))
    assert os.path.exists(os.path.join(workdir, 'square:process:0.log'))
    assert os.path.exists(os.path.join(workdir, 'square:merge.log'))


def test_psyrun_workdir_contents_load_balanced(taskenv):
    workdir = os.path.join('psy-work', 'square_load_balanced')
    os.remove(os.path.join(taskenv.taskdir, 'psy-conf.py'))
    psy_main(['run', '--taskdir', taskenv.taskdir, 'square_load_balanced'])
    assert os.path.exists(os.path.join(workdir, 'in.pkl'))
    assert os.path.exists(os.path.join(workdir, 'result.pkl'))
    assert os.path.exists(os.path.join(
        workdir, 'square_load_balanced:pspace.py'))
    assert os.path.exists(os.path.join(
        workdir, 'square_load_balanced:process.py'))
    assert os.path.exists(os.path.join(
        workdir, 'square_load_balanced:process:0.log'))
    assert os.path.exists(os.path.join(
        workdir, 'square_load_balanced:process:1.log'))


def test_psyrun_file_dep(taskenv):
    with open(os.path.join(taskenv.taskdir, 'in.txt'), 'w') as f:
        f.write('2')
    psy_main(['run', '--taskdir', taskenv.taskdir, 'file_dep'])
    result = PickleStore().load(os.path.join(
        taskenv.workdir, 'file_dep', 'result.pkl'))
    assert sorted(result['y']) == [4]

    # Ensure that modification time changes as some file systems only support
    # 1s resolution.
    time.sleep(1)

    with open(os.path.join(taskenv.taskdir, 'in.txt'), 'w') as f:
        f.write('3')
    psy_main(['run', '--taskdir', taskenv.taskdir, 'file_dep'])
    result = PickleStore().load(os.path.join(
        taskenv.workdir, 'file_dep', 'result.pkl'))
    assert sorted(result['y']) == [8]


@pytest.mark.filterwarnings(JobsRunningWarning)
def test_psyrun_does_not_resubmit_queued_jobs(taskenv, scheduler):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    expected = scheduler.joblist
    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    assert len(expected) == len(scheduler.joblist)
    assert all(x['id'] == y['id'] for x, y in zip(expected, scheduler.joblist))


def test_psyrun_remerges_if_result_is_missing(taskenv, scheduler):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    scheduler.consume()
    os.remove(os.path.join(taskenv.workdir, 'mocked_scheduler', 'result.pkl'))

    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    assert len(scheduler.joblist) == 1
    assert 'merge' in scheduler.joblist[0]['name']


def test_psyrun_no_resubmits_if_result_is_uptodate(taskenv, scheduler):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    scheduler.consume()
    for dirpath, _, filenames in os.walk(taskenv.workdir):
        for filename in filenames:
            if filename == 'result.pkl':
                continue
            os.remove(os.path.join(dirpath, filename))

    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    assert len(scheduler.joblist) == 0


def test_psyrun_resubmits_for_missing_job_output(taskenv, scheduler):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    scheduler.consume()
    os.remove(os.path.join(taskenv.workdir, 'mocked_scheduler', 'result.pkl'))
    os.remove(os.path.join(
        taskenv.workdir, 'mocked_scheduler', 'out', '0.pkl'))

    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    assert len(scheduler.joblist) == 2
    assert 'process:0' in scheduler.joblist[0]['name']
    assert 'merge' in scheduler.joblist[1]['name']


def test_psyrun_does_not_resubmit_split_if_infiles_uptodate(
        taskenv, scheduler):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    scheduler.consume()
    for dirpath, _, filenames in os.walk(taskenv.workdir):
        for filename in filenames:
            if os.path.basename(dirpath) == 'out':
                continue
            os.remove(os.path.join(dirpath, filename))

    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    for job in scheduler.joblist:
        assert 'split' not in job['name']


def test_psyrun_resubmits_jobs_if_necessary(taskenv, scheduler):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    scheduler.consume()
    shutil.rmtree(taskenv.workdir)

    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    assert len(scheduler.joblist) == 6
    assert 'split' in scheduler.joblist[0]['name']
    for i in range(4):
        assert 'process:{0}'.format(i) in scheduler.joblist[i + 1]['name']
    assert 'merge' in scheduler.joblist[5]['name']


def test_psyrun_shows_error_if_resubmit_of_queued_job_necessary(
        taskenv, scheduler):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    scheduler.consume_job(scheduler.joblist[0])
    scheduler.consume_job(scheduler.joblist[1])
    expected = scheduler.joblist
    time.sleep(1)
    t = time.time()
    os.utime(
        os.path.join(taskenv.taskdir, 'task_mocked_scheduler.py'),
        (t, t))

    with pytest.warns(JobsRunningWarning):
        psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    assert len(expected) == len(scheduler.joblist)
    assert all(x['id'] == y['id'] for x, y in zip(expected, scheduler.joblist))


def test_psyrun_resubmits_merge_if_result_is_outdated(taskenv, scheduler):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    scheduler.consume()
    time.sleep(1)
    t = time.time()
    os.utime(
        os.path.join(taskenv.taskdir, 'task_mocked_scheduler.py'),
        (t, t))
    for i in range(4):
        os.utime(os.path.join(
            taskenv.workdir, 'mocked_scheduler', 'in',
            str(i) + '.pkl'), (t, t))
        os.utime(os.path.join(
            taskenv.workdir, 'mocked_scheduler', 'out',
            str(i) + '.pkl'), (t, t))

    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    assert len(scheduler.joblist) == 1
    assert 'merge' in scheduler.joblist[0]['name']


def test_psyrun_resubmits_process_and_merge_if_outfile_is_outdated(
        taskenv, scheduler):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    scheduler.consume()
    time.sleep(1)
    t = time.time()
    os.utime(
        os.path.join(taskenv.taskdir, 'task_mocked_scheduler.py'),
        (t, t))
    for i in range(4):
        os.utime(os.path.join(
            taskenv.workdir, 'mocked_scheduler', 'in',
            str(i) + '.pkl'), (t, t))

    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    assert len(scheduler.joblist) == 5
    for i in range(4):
        assert 'process:{0}'.format(i) in scheduler.joblist[i]['name']
    assert 'merge' in scheduler.joblist[4]['name']


def test_psyrun_resubmits_all_if_infile_is_outdated(
        taskenv, scheduler):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    scheduler.consume()
    time.sleep(1)
    t = time.time()
    os.utime(
        os.path.join(taskenv.taskdir, 'task_mocked_scheduler.py'),
        (t, t))

    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    assert len(scheduler.joblist) == 6
    assert 'split' in scheduler.joblist[0]['name']
    for i in range(4):
        assert 'process:{0}'.format(i) in scheduler.joblist[i + 1]['name']
    assert 'merge' in scheduler.joblist[5]['name']


def test_multiple_splits(taskenv):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'square2'])
    result = PickleStore().load(os.path.join(
        taskenv.workdir, 'square2', 'result.pkl'))
    assert sorted(result['y']) == [0, 1, 4, 9]


def test_psy_run_runs_all_tasks(taskenv):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'square', 'square2'])
    result = PickleStore().load(
        os.path.join(taskenv.workdir, 'square', 'result.pkl'))
    assert sorted(result['y']) == [0, 1, 4, 9]
    result = PickleStore().load(os.path.join(
        taskenv.workdir, 'square2', 'result.pkl'))
    assert sorted(result['y']) == [0, 1, 4, 9]


def test_psy_list(taskenv, capsys):
    expected = set()
    for entry in os.listdir(taskenv.taskdir):
        m = re.match(r'^task_(.*)\.py$', entry)
        if m:
            expected.add(m.group(1))

    psy_main(['list', '--taskdir', taskenv.taskdir])
    out, _ = capsys.readouterr()
    listed = {x.strip() for x in out.split('\n') if x != ''}
    assert listed == expected


def test_psy_status(taskenv, capsys):
    psy_main(['status', '--taskdir', taskenv.taskdir, 'square'])
    out, _ = capsys.readouterr()
    assert out == """square:
  0 out of 4 rows completed.

"""

    psy_main(['status', '--taskdir', taskenv.taskdir, '-v', 'square'])
    out, _ = capsys.readouterr()
    assert out == """square:
  0 out of 4 rows completed.
  Missing parameter sets:
    {'x': 0}
    {'x': 1}
    {'x': 2}
    {'x': 3}
  Queued parameter sets:
  No failed jobs.

"""

    psy_main(['run', '--taskdir', taskenv.taskdir, 'square'])
    out, _ = capsys.readouterr()
    psy_main(['status', '--taskdir', taskenv.taskdir, '-v', 'square'])
    out, _ = capsys.readouterr()
    assert out == """square:
  4 out of 4 rows completed.
  Missing parameter sets:
  Queued parameter sets:
  No failed jobs.

"""

    os.remove(os.path.join(taskenv.workdir, 'square', 'out', '2.pkl'))
    psy_main(['status', '--taskdir', taskenv.taskdir, '-v', 'square'])
    out, _ = capsys.readouterr()
    assert out == """square:
  4 out of 4 rows completed.
  Missing parameter sets:
  Queued parameter sets:
  Failed jobs:
    square:process:2

"""

    os.remove(os.path.join(taskenv.workdir, 'square', 'result.pkl'))
    psy_main(['status', '--taskdir', taskenv.taskdir, '-v', 'square'])
    out, _ = capsys.readouterr()
    assert out == """square:
  3 out of 4 rows completed.
  Missing parameter sets:
    {'x': 2}
  Queued parameter sets:
  Failed jobs:
    square:process:2

"""


def test_psy_status_load_balanced(taskenv, capsys):
    psy_main(['status', '--taskdir', taskenv.taskdir, 'square_load_balanced'])
    out, _ = capsys.readouterr()
    assert out == """square_load_balanced:
  0 out of 4 rows completed.

"""

    psy_main(['status', '--taskdir', taskenv.taskdir, '-v',
              'square_load_balanced'])
    out, _ = capsys.readouterr()
    assert out == """square_load_balanced:
  0 out of 4 rows completed.
  Missing parameter sets:
    {'x': 0}
    {'x': 1}
    {'x': 2}
    {'x': 3}

"""

    psy_main(['run', '--taskdir', taskenv.taskdir, 'square_load_balanced'])
    out, _ = capsys.readouterr()
    psy_main(['status', '--taskdir', taskenv.taskdir, '-v',
              'square_load_balanced'])
    out, _ = capsys.readouterr()
    assert out == """square_load_balanced:
  4 out of 4 rows completed.
  Missing parameter sets:

"""


def test_psy_kill(taskenv, scheduler):
    psy_main(['run', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    assert len(scheduler.joblist) > 0
    psy_main(['kill', '--taskdir', taskenv.taskdir, 'mocked_scheduler'])
    assert len(scheduler.joblist) == 0


@pytest.mark.parametrize('store', [PickleStore(), NpzStore(), H5Store()])
def test_psy_merge(tmpdir, store):
    resultfile = os.path.join(str(tmpdir), 'result' + store.ext)
    outdir = os.path.join(str(tmpdir), 'out')
    if not os.path.exists(outdir):
        os.mkdir(outdir)

    data_segments = ({'x': [1]}, {'x': [2]})
    for i, d in enumerate(data_segments):
        store.save(os.path.join(outdir, str(i) + store.ext), d)

    psy_main(['merge', outdir, resultfile])
    merged = store.load(resultfile)
    assert list(merged.keys()) == ['x']
    assert np.all(np.asarray(sorted(merged['x'])) == np.array([1, 2]))


def test_new_task(taskenv):
    psy_main(['new-task', '--taskdir', taskenv.taskdir, 'new-task-test'])
    assert os.path.exists(
        os.path.join(taskenv.taskdir, 'task_new-task-test.py'))
    assert psy_main(
        ['new-task', '--taskdir', taskenv.taskdir, 'new-task-test']) != 0


@pytest.mark.parametrize('scheduler_str', ['psyrun.scheduler.Sqsub', 'Sqsub'])
def test_new_task_scheduler_arg(taskenv, scheduler_str):
    psy_main(['new-task', '--taskdir', taskenv.taskdir, '-s', scheduler_str,
              'new-task-test'])
    path = os.path.join(taskenv.taskdir, 'task_new-task-test.py')
    with open(path, 'r') as f:
        data = f.read()
    assert 'from psyrun.scheduler import Sqsub' in data
    assert 'scheduler = Sqsub' in data
    assert 'scheduler_args = {' in data
    assert 'scheduler_args = {}' not in data



"""Base store interface."""

import os.path
from pkg_resources import iter_entry_points


class Store(object):
    """Defines the interface of stores.

    Register implemented stores as
    `entry points <https://setuptools.readthedocs.io/en/latest/setuptools.html#dynamic-discovery-of-services-and-plugins>`_
    in the ``psyrun.stores`` groupn. For example, add the following to the
    ``setup`` call in your store's ``setup.py`` for a store providing the
    ``.ext`` format::

        entry_points={
            'psyrun.stores': ['.ext = pkg.name:ClassName'],
        }

    Attributes
    ----------
    ext : str
        Filename extension used by the store.
    """

    ext = ''

    def append(self, filename, data):
        """Append data to file.

        When trying to append data to a non-existing file, a new file will be
        created. The backend may require that a file was created with this
        function to be able to append to it.

        Parameters
        ----------
        filename : str
            Filename of file to append the data to.
        data : dict
            Dictionary with data to append.
        """
        raise NotImplementedError()

    def save(self, filename, data):
        """Save data to a file.

        Parameters
        ----------
        filename : str
            Filename of file to save data to.
        data : dict
            Dictionary with data to store.
        """
        raise NotImplementedError()

    def load(self, filename, row=None):
        """Load data from a file.

        Parameters
        ----------
        filename : str
            Filename of file to load data from.
        row : int, optional
            If given, only the row with this index will be loaded.

        Returns
        -------
        dict
            Loaded data.
        """
        raise NotImplementedError()


def _safe_ep_load():
    for ep in iter_entry_points('psyrun.stores'):
        try:
            yield ep.name, ep.load()
        except ImportError:
            pass


class AutodetectStore(Store):
    """Automatically selects the store based on the file extension."""


    registry = dict(_safe_ep_load())

    @classmethod
    def get_concrete_store(cls, filename):
        _, ext = os.path.splitext(filename)
        return cls.registry[ext.lower()]()

    def append(self, filename, data):
        return self.get_concrete_store(filename).append(filename, data)

    def save(self, filename, data):
        return self.get_concrete_store(filename).save(filename, data)

    def load(self, filename, row=None):
        return self.get_concrete_store(filename).load(filename, row=row)



from psyrun import Param


pspace = Param(x=range(4))
overwrite_dirty = False
exclude_from_result = ['z']


def setup(proc_id):
    assert 0 <= proc_id < 4
    return {'p': 2}


def execute(x, p):
    return {'y': x ** p, 'z': -1}



import pytest

from psyrun.backend.load_balancing import LoadBalancingWorker
from psyrun.pspace import Param
from psyrun.store.h5 import H5Store
from psyrun.store.npz import NpzStore
from psyrun.store.pickle import PickleStore


def square(a, p):
    return {'x': a ** p}


def setup_fn(i):
    assert i == 0
    return {'p': 2}


@pytest.mark.parametrize('store', [PickleStore(), NpzStore(), H5Store()])
def test_load_balancing_worker(tmpdir, store):
    infile = str(tmpdir.join('in.npz'))
    outfile = str(tmpdir.join('out.npz'))
    statusfile = str(tmpdir.join('status'))
    store.save(infile, Param(a=range(7)).build())
    LoadBalancingWorker.create_statusfile(statusfile)
    worker = LoadBalancingWorker(0, infile, outfile, statusfile, store)
    worker.start(square, setup_fn)
    result = store.load(outfile)
    assert sorted(result['a']) == sorted(range(7))
    assert sorted(result['x']) == [i ** 2 for i in range(7)]



from psyrun import Param
from psyrun.backend import LoadBalancingBackend


pspace = Param(x=range(4))
max_jobs = 2
backend = LoadBalancingBackend
exclude_from_result = ['z']


def setup(proc_id):
    assert proc_id in (0, 1)
    return {'p': 2}


def execute(x, p):
    return {'y': x ** p, 'z': -1}
